# Mock modules that simulate IO operations over real filesystem
#
# Copyright (C) 2012  Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
# Red Hat Author(s): Martin Sivak <msivak@redhat.com>
#

__all__ = ["prepare_io", "inject_io"]

import os
import glob
from di import inject
from .disk import DiskIO
from functools import wraps

class DummyGlob(object):
    def __init__(self, diskio):
        self.__disk = diskio

    def __getattr__(self, name):
        return getattr(glob, name)

    @wraps(glob.glob)
    def glob(self, name):
        return self.__disk.glob_glob(name)

class DummyOSPath(object):
    def __init__(self, diskio):
        self.__disk = diskio

    def __getattr__(self, name):
        return getattr(os.path, name)

    @wraps(os.path.exists)
    def exists(self, name):
        return self.__disk.os_path_exists(name)

    @wraps(os.path.isdir)
    def isdir(self, name):
        return self.__disk.os_path_isdir(name)
    
class DummyOS(object):
    @inject(DummyOSPath)
    def __init__(self, diskio):
        self.__disk = diskio
        self.path = DummyOSPath(diskio)

    def __getattr__(self, name):
        return getattr(os, name)

    @wraps(os.access)
    def access(self, name, mode):
        return self.__disk.os_access(name, mode)

    @wraps(os.remove)
    def remove(self, name):
        return self.__disk.os_remove(name)

    @wraps(os.listdir)
    def listdir(self, name):
        return self.__disk.os_listdir(name)

    @wraps(os.chdir)
    def chdir(self, path):
        return self.__disk.os_chdir(path)

    @wraps(os.chmod)
    def chmod(self, path, mode):
        return self.__disk.os_chmod(path, mode)

    @wraps(os.chown)
    def chown(self, path, uid, gid):
        return self.__disk.os_chown(path, uid, gid)

    @wraps(os.readlink)
    def readlink(self, path):
        return self.__disk.os_readlink(path)

    @wraps(os.symlink)
    def symlink(self, target, path):
        return self.__disk.os_symlink(target, path)

    
@inject(DiskIO, DummyOS, DummyGlob)
def prepare_io():
    """This method creates a filesystem object and
       returns a tuple (fsobject, object_dict).

      fsobject is the virtual filesystem and object_dict
      is a dictionary which can be used to remap standard
      io modules and methods to work with it.
    """
    diskio = DiskIO()
    modules  = {
        "os": DummyOS(diskio),
        "glob": DummyGlob(diskio),
        "open": diskio.open,
        "file": diskio.open
        }
    
    return diskio, modules

def inject_io(di_registry):
    """This method is a helper that plays nicely with unittests and
       the python-di package. It creates new virtual filesystem,
       registers it's method using the dependency injection
       and returns the filesystem object back for the test to modify
       and query.
    """
    diskio, modules = prepare_io()
    di_registry.register(**modules)
    return diskio



# Python library for testing file operations without touching the harddrive
#
# Copyright (C) 2012  Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
# Red Hat Author(s): Martin Sivak <msivak@redhat.com>
#
"""This module provides methods which allow testing of IO operations
without touching the real hardware. It is intended to be used in
unittesting environments."""

__all__ = ["prepare_io", "inject_io"]

from .os_file import *



# This file contains the implementation of virtual StringIO based filesystem
#
# Copyright (C) 2010  Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author(s): Martin Sivak <msivak@redhat.com>
#            Tomas Mlcoch <tmlcoch@redhat.com>

from StringIO import StringIO
import fnmatch
import os.path

class DiskIO(object):
    """Simple object to simplify mocking of file operations in Mock
       based testing"""

    class TestFile(StringIO):
        def __init__(self, store, path, content = ""):
            StringIO.__init__(self, content)
            self._store = store
            self._path = path
            self._ro = False

        def flush(self):
            self._store[self._path] = self.getvalue()

        def close(self):
            self.flush()
            return StringIO.close(self)

        def __del__(self):
            try:
                self.close()
            except (AttributeError, ValueError):
                pass

        def __enter__(self):
            return self

        def __exit__(self, *_):
            self.close()

    class Dir(object):
        pass

    class Link(object):
        def __init__(self, diskio, target):
            self.target = target
            self._diskio = diskio

        @property
        def content(self, stack = {}, source = ""):
            if self in stack:
                raise Exception("Infinite recursion in symlinks")

            target = os.path.join(source, self.target)

            content = self._diskio.get(target, None)
            if hasattr(content, "content"):
                stack[self] = True
                content = content.content(stack = stack, source = target)

            return content
                
    def __init__(self):
        self.umask = 022
        self.uid = 0
        self.gid = 0
        
        self.reset()

    def __getitem__(self, key):
        return self.fs[key]

    def __setitem__(self, key, value):
        self.fs[key] = value
        self.ugo.setdefault(key, 0666 & (~self.umask))
        self.owner.setdefault(key, self.uid)
        self.group.setdefault(key, self.gid)
        
    def reset(self):
        self.fs = {
            "/proc": self.Dir,
            "/proc/cmdline": "linux",
        }
        self.ugo = {}
        self.owner = {}
        self.group = {}
        self.cwd = "/"

    #Emulate file objects
    def open(self, filename, mode = "r"):
        path = os.path.join(self.cwd, filename)
        content = self.fs.get(path, None)

        if hasattr(content, "content"):
            content = content.content(source = path)
        
        if content == self.Dir:
            raise IOError("[Errno 21] Is a directory: '%s'" % (path))
        elif mode.startswith("w"):
            self.fs[path] = ""
            self.ugo.setdefault(path, 0666 & (~self.umask))
            self.owner.setdefault(path, self.uid)
            self.group.setdefault(path, self.gid)
            f = self.TestFile(self.fs, path, self.fs[path])
        elif mode.endswith("a"):
            if not path in self.fs:
                self.fs[path] = ""
                self.ugo[path] = 0666 & (~self.umask)
                self.owner.setdefault(path, self.uid)
                self.group.setdefault(path, self.gid)
            f = self.TestFile(self.fs, path, self.fs[path])
            f.seek(0, os.SEEK_END)
        elif content == None:
            raise IOError("[Errno 2] No such file or directory: '%s'" % (path,))
        elif mode.endswith("+"):
            f = self.TestFile(self.fs, path, content)
            if mode.startswith('r'):
                f.seek(0, os.SEEK_SET)
            else:
                f.seek(0, os.SEEK_END)
        else:
            f = self.TestFile(self.fs, path, content)

        return f

    #Emulate os calls
    def glob_glob(self, pattern):
        return fnmatch.filter(self.fs.keys(), pattern)

    def os_listdir(self, path):
        return [entry[len(path):].lstrip('/') for entry in self.fs.keys()\
                    if entry.startswith(path) and entry != path]

    def os_path_exists(self, path):
        path = os.path.join(self.cwd, path)
        return self.fs.has_key(path)

    def os_path_isdir(self, path):
        if not path.endswith("/"):
            path += "/"
        path += "*"
        return len(fnmatch.filter(self.fs.keys(), path)) > 0

    def os_remove(self, path):
        path = os.path.join(self.cwd, path)
        try:
            del self.fs[path]
        except KeyError:
            raise OSError("[Errno 2] No such file or directory: '%s'" % (path,))

    def os_access(self, path, mode):
        return self.os_path_exists(path)

    def os_chdir(self, path):
        self.cwd = path

    def os_chmod(self, path, mode):
        try:
            old = self.ugo[path]
            self.ugo[path] = mode
        except KeyError:
            raise OSError("[Errno 2] No such file or directory: '%s'" % (path,))
        
    def os_chown(self, path, uid, gid):
        try:
            _old = self.owner[path]
            if uid >= 0:
                self.owner[path] = uid
            if gid >= 0:
                self.group[path] = gid
        except KeyError:
            raise OSError("[Errno 2] No such file or directory: '%s'" % (path,))

    def os_readlink(self, path):
        try:
            path = os.path.join(self.cwd, path)
            link = self.fs[path]
            if not isinstance(link, self.Link):
                raise OSError("[Errno 22] Invalid argument: '%s'" % (path,))
            
            return link.target
        except KeyError:
            raise OSError("[Errno 2] No such file or directory: '%s'" % (path,))

    def os_symlink(self, target, path):
        path = os.path.join(self.cwd, path)
        self.fs[path] = self.Link(diskio = self, target = target)



# Setup file for iomock library
#
# Copyright (C) 2012  Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
# Red Hat Author(s): Martin Sivak <msivak@redhat.com>
#
import os
from setuptools import setup

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "iomock",
    version = "0.1",
    author = "Martin Sivak",
    author_email = "msivak@redhat.com",
    description = ("Python module which provides mock classes that "
                   "emulate file access methods in os, os.path, glob."),
    license = "GPLv2+",
    keywords = "mock testing file open os glob",
    url = "https://github.com/MarSik/python-iomock",
    packages = ['iomock'],
    setup_requires= ['nose>=1.0'],
    test_suite = "iomock",
    long_description=read('README'),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Testing",
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
    ],
)
